#!/usr/bin/env python3

import os
from pathlib import Path
from typing import List, Optional, Sequence

from pydantic import Field, field_validator

from avtdl.core import utils
from avtdl.core.config import Plugins
from avtdl.core.interfaces import Actor, ActorConfig, ActorEntity, Event, EventType, Record, TextRecord
from avtdl.core.monitors import HIGHEST_UPDATE_INTERVAL, TaskMonitor, TaskMonitorEntity
from avtdl.core.utils import Fmt, OutputFormat, read_file, sanitize_filename

Plugins.register('from_file', Plugins.kind.ASSOCIATED_RECORD)(TextRecord)


@Plugins.register('from_file', Plugins.kind.ACTOR_CONFIG)
class FileMonitorConfig(ActorConfig):
    pass


@Plugins.register('from_file', Plugins.kind.ACTOR_ENTITY)
class FileMonitorEntity(TaskMonitorEntity):
    encoding: Optional[str] = None
    """encoding used to open the monitored file. If not specified, default system-wide encoding is used"""
    path: Path
    """path to the monitored file"""
    split_lines: bool = False
    """if true, each line of the file will create a separate record. Otherwise, a single record will be generated with the entire file content"""
    update_interval: float = 60
    """how often the monitored file should be checked, in seconds"""
    mtime: float = Field(exclude=True, default=-1)
    """internal variable to persist state between updates. Used to check if the file has changed"""
    base_update_interval: float = Field(exclude=True, default=60)
    """internal variable to persist state between updates. Used to restore configured update interval after delay on network request error"""


    def __post_init__(self):
        self.path = Path(self.path)
        self.base_update_interval = self.update_interval


@Plugins.register('from_file', Plugins.kind.ACTOR)
class FileMonitor(TaskMonitor):
    """
    Monitor content of a text file

    On specified intervals, check existence and last modification time
    of target file, and if it has changed, read file contents
    either line by line or as a whole and emit it as a text record(s).

    Records are not checked for uniqueness, so appending content to the end
    of the existing file will produce duplicates of already sent records.
    """

    async def get_new_records(self, entity: FileMonitorEntity) -> Sequence[TextRecord]:
        if not self.has_changed(entity):
            return []
        try:
            records = self.get_records(entity)
            if entity.update_interval != entity.base_update_interval:
                entity.update_interval = entity.base_update_interval
            return records
        except Exception as e:
            self.logger.warning(f'[{entity.name}] error when processing file "{entity.path}": {e}')
            entity.update_interval = max(entity.update_interval * 1.2, HIGHEST_UPDATE_INTERVAL)
            return []

    def exists(self, entity: FileMonitorEntity) -> bool:
        if not entity.path.exists():
            entity.mtime = -1
            return False
        else:
            return True

    def has_changed(self, entity: FileMonitorEntity) -> bool:
        if not self.exists(entity):
            return False
        try:
            current_mtime = os.stat(entity.path).st_mtime
        except OSError as e:
            self.logger.debug(f'[{entity.name}] failed to get file info for "{entity.path}": {e}')
            return False
        if current_mtime == entity.mtime:
            return False
        else:
            entity.mtime = current_mtime
            return True

    def get_records(self, entity: FileMonitorEntity) -> List[TextRecord]:
        records = []
        if self.exists(entity):
            text = read_file(entity.path)
            if entity.split_lines:
                lines = text.split('\n')
            else:
                lines = [text]
            for line in lines:
                record = TextRecord(text=line.strip())
                records.append(record)
        return records


Plugins.register('to_file', Plugins.kind.ASSOCIATED_RECORD)(Event)


@Plugins.register('to_file', Plugins.kind.ACTOR_CONFIG)
class FileActionConfig(ActorConfig):
    pass


@Plugins.register('to_file', Plugins.kind.ACTOR_ENTITY)
class FileActionEntity(ActorEntity):
    path: Optional[Path] = None
    """directory where output file should be created. Default is current directory"""
    filename: str
    """name of the output file. Supports templating with `{...}`"""
    encoding: Optional[str] = 'utf8'
    """output file encoding"""
    output_format: OutputFormat = Field(default=OutputFormat.str, description='one of `' + "`, `".join(OutputFormat.__members__) + '`')
    """should record be written in output file as plain text or json"""
    overwrite: bool = True
    """whether file should be overwritten in if it already exists"""
    append: bool = True
    """if true, new record will be written at the end of the file without overwriting already present lines"""
    prefix: str = ''
    """string that will be appended before the record text. Can be used to separate records from each other or for simple templating"""
    postfix: str = '\n'
    """string that will be appended after the record text"""

    @field_validator('path')
    @classmethod
    def check_dir(cls, path: Optional[Path]) -> Path:
        if path is None:
            return Path.cwd()
        if utils.check_dir(path):
            return path
        raise ValueError(f'check if provided path points to a writeable directory')


@Plugins.register('to_file', Plugins.kind.ACTOR)
class FileAction(Actor):
    """
    Write record to a text file

    Takes a record coming from a Chain, converts it to text representation,
    and writes to a file in given directory. When a file already exists,
    new records can be appended to the end of the file or overwrite it.

    Output file name can be static or generated dynamically based on the template
    filled with values from the record fields: every occurrence of `{text}`
    in filename will be replaced with the value of the `text` field of processed
    record, if the record has one.

    Allows writing the record as human-readable text representation or as names and
    values of the record fields in json format. For custom format template, pass the record
    through `filter.format` plugin prior to this one.

    Produces `Event` with `error` type if writing to target file fails.

    Note discrepancy between default value of `encoding` setting between `from_file`
    and `to_file` plugins. Former is expected to be able to read files produced by
    different software and therefore relies on system-wide settings. It would make
    sense to do the same in the latter, but it would introduce possibility of failing
    to write records containing text with Unicode codepoints that cannot be represented
    using system-wide encoding.
    """

    def handle(self, entity: FileActionEntity, record: Record):
        filename = Fmt.format(entity.filename, record)
        filename = sanitize_filename(filename)
        if entity.path is None:
            path = Path.cwd().joinpath(filename)
        else:
            path = Path(entity.path).joinpath(filename)
        if path.exists() and not entity.overwrite:
            self.logger.debug(f'[{entity.name}] file {path} already exists, not overwriting')
            return
        mode = 'at' if entity.append else 'wt'
        try:
            text = entity.prefix + Fmt.save_as(record, entity.output_format) + entity.postfix
            with open(path, mode, encoding=entity.encoding) as fp:
                fp.write(text)
        except Exception as e:
            message = f'error in {self.conf.name}.{entity}: {e}'
            self.on_record(entity, Event(event_type=EventType.error, text=message))
            self.logger.exception(message)



#!/usr/bin/env python3

import argparse
import asyncio
import logging
import os
from pathlib import Path

import yaml
from pydantic import ValidationError

from avtdl.core.config import ConfigParser
from avtdl.core.info import generate_plugins_description
from avtdl.core.loggers import set_logging_format, silence_library_loggers
from avtdl.core.utils import read_file


def load_config(path: Path):
    if not os.path.exists(path):
        print('Configuration file {} does not exist'.format(path))
        raise SystemExit
    try:
        config_text = read_file(path)
        config = yaml.load(config_text, Loader=yaml.FullLoader)
    except Exception as e:
        print('Failed to parse configuration file:')
        print(e)
        raise SystemExit from e
    return config


def parse_config(conf):
    try:
        actors, chains = ConfigParser.parse(conf)
    except ValidationError as e:
        logging.error(e)
        raise SystemExit from e
    except Exception as e:
        logging.exception(e)
        raise SystemExit from e
    return actors, chains


def handler(loop, context):
    logging.exception(f'unhandled exception in event loop:', exc_info=context.get('exception'))
    loop.default_exception_handler(context)


async def run(runnables):
    loop = asyncio.get_running_loop()
    loop.set_exception_handler(handler)
    loop.slow_callback_duration = 100

    tasks = []
    for runnable in runnables:
        task = asyncio.create_task(runnable.run(), name=runnable.__class__.__name__)
        tasks.append(task)
    while True:
        done, pending = await asyncio.wait(tasks, return_when=asyncio.FIRST_EXCEPTION)
        for task in done:
            if not task.done():
                continue
            if task.exception() is not None:
                logging.warning(f'task {task.get_name()} has terminated with exception', exc_info=task.exception())
        if not pending:
            break
        tasks = pending
    logging.info('all tasks are finished in the main loop')


def start(args):
    conf = load_config(args.config)
    actors, chains = parse_config(conf)

    for chain_name, chain_instance in chains.items():
        for actor_name, entities in chain_instance.conf:
            actor = actors.get(actor_name)
            if actor is None:
                logging.warning(f'chain "{chain_name}" references actor "{actor_name}, absent in "Actors" section. It might be a typo in the chain configuration')
                continue
            orphans = set(entities) - actor.entities.keys()
            for orphan in orphans:
                logging.warning(f'chain "{chain_name}" references "{actor_name}: {orphan}", but actor "{actor_name}" has no "{orphan}" entity. It might be a typo in the chain conf configuration')

    asyncio.run(run(actors.values()), debug=True)


def make_docs(args):
    output = args.plugins_doc
    doc = generate_plugins_description(output.suffix == '.html')
    try:
        with open(output, 'wt', encoding='utf8') as fp:
            fp.write(doc)
    except OSError as e:
        logging.error(f'failed to write documentation file "{output}": {e}')
        raise SystemExit from e
    

def main():
    description = '''Tool for monitoring rss feeds and other sources and running commands for new entries'''
    parser = argparse.ArgumentParser(description=description)
    help_v = 'set loglevel to DEBUG'
    parser.add_argument('-d', '--debug', action='store_true', default=False, help=help_v)
    help_v = 'set loglevel to INFO'
    parser.add_argument('-v', '--verbose', action='store_true', default=False, help=help_v)
    help_c = 'specify path to configuration file to use instead of default'
    parser.add_argument('-c', '--config', type=Path, default='config.yml', help=help_c)
    help_h = 'write plugins documentation in given file and exit. Documentation format is deduced by file extension: html document for ".html", markdown otherwise'
    parser.add_argument('-p', '--plugins-doc', type=Path, required=False, help=help_h)
    args = parser.parse_args()

    if args.debug:
        log_level = logging.DEBUG
    elif args.verbose:
        log_level = logging.INFO
    else:
        log_level = logging.WARNING

    set_logging_format(log_level)
    silence_library_loggers()

    if args.plugins_doc is not None:
        make_docs(args)
    else:
        start(args)


if __name__ == "__main__":
    main()



import json
from typing import Any, List, Optional, Sequence, Tuple, Union

import aiohttp

from avtdl.core import utils
from avtdl.core.interfaces import MAX_REPR_LEN, Record
from avtdl.core.monitors import PagedFeedMonitor, PagedFeedMonitorConfig, PagedFeedMonitorEntity
from avtdl.core.plugins import Plugins
from avtdl.plugins.youtube.common import get_continuation_token, get_initial_data, get_innertube_context, handle_consent, prepare_next_page_request, thumbnail_url, \
    video_url
from avtdl.plugins.youtube.community_info import CommunityPostInfo, SharedCommunityPostInfo, get_posts_renderers, get_renderers, get_shared_posts_renderers


@Plugins.register('community', Plugins.kind.ASSOCIATED_RECORD)
class CommunityPostRecord(Record, CommunityPostInfo):
    """Youtube community post content"""
    channel_id: str
    """channel ID in old format"""
    post_id: str
    """unique id of the post"""
    author: str
    """author's channel name"""
    avatar_url: Optional[str] = None
    """link to the avatar of the channel"""
    vote_count: str
    """current number of upvotes"""
    sponsor_only: bool
    """indicates whether the post is member-only"""
    published_text: str
    """localized text saying how long ago the video was uploaded"""
    full_text: str
    """post contents as plaintext"""
    attachments: List[str]
    """list of links to attached images or video thumbnails"""
    video_id: Optional[str] = None
    """if the post links to youtube video will contain video id, otherwise absent"""
    original_post: Optional['CommunityPostRecord'] = None
    """for reposts contains original post content, otherwise absent"""

    def __repr__(self) -> str:
        text = self.full_text.replace('\n', ' • ')[:MAX_REPR_LEN]
        return f'{self.post_id} [{self.author}] {text}'

    def __str__(self) -> str:
        channel_post_url = f'https://www.youtube.com/channel/{self.channel_id}/community?lb={self.post_id}'
        sponsor_only = '(Member only)' if self.sponsor_only else ''

        header = f'[{self.author}, {self.published_text} {sponsor_only}] {self.vote_count}'
        body = self.full_text
        attachments = '\n'.join(self.attachments)
        video = ''
        if self.video_id and self.full_text.find(self.video_id) == -1:
            video = video_url(self.video_id)
        original_post = str(self.original_post) if self.original_post else ''
        return '\n'.join((channel_post_url, header, body, video, attachments, original_post))

    def discord_embed(self) -> List[dict]:
        channel_url = f'https://www.youtube.com/channel/{self.channel_id}'
        post_url = f'https://www.youtube.com/post/{self.post_id}'

        attachments = '\n'.join(self.attachments) if len(self.attachments) > 4 else ''
        video = ''
        if self.video_id and self.full_text.find(self.video_id) == -1:
            video = video_url(self.video_id)
        original_post = str(self.original_post) if self.original_post else ''
        text = '\n'.join([self.full_text, attachments, video, original_post])

        embed = {
            'title': self.post_id,
            'description': text,
            'url': post_url,
            'color': None,
            'author': {'name': self.author, 'url': channel_url, 'icon_url': self.avatar_url},
            'footer': {'text': self.published_text}
        }
        if self.sponsor_only:
            embed['fields'] = [{'name': 'Member only', 'value': ''}]
        if self.video_id:
            embed['image'] = {'url': thumbnail_url(self.video_id)}
        if self.attachments:
            images: List[dict] = [{'url': post_url, 'image': {'url': attachment}} for attachment in self.attachments]
            if embed.get('image') is None:
                embed['image'] = images.pop(0)['image']
            embeds = [embed, *images]
        else:
            embeds = [embed]
        return embeds

@Plugins.register('community', Plugins.kind.ASSOCIATED_RECORD)
class SharedCommunityPostRecord(Record, SharedCommunityPostInfo):
    """Youtube community post that is itself a repost of another post"""
    channel_id: str
    """channel ID in old format"""
    post_id: str
    """unique id of the post"""
    author: str
    """author's channel name"""
    avatar_url: Optional[str] = None
    """link to the avatar of the channel"""
    published_text: str
    """localized text saying how long ago the video was uploaded"""
    full_text: str
    """post content"""
    original_post: Optional['CommunityPostRecord'] = None
    """not present in shared post"""

    def __repr__(self) -> str:
        text = self.full_text.replace('\n', ' • ')[:MAX_REPR_LEN]
        return f'{self.post_id} [{self.author}] {text}'

    def __str__(self) -> str:
        channel_post_url = f'https://www.youtube.com/channel/{self.channel_id}/community?lb={self.post_id}'

        header = f'[{self.author}, {self.published_text} ]'
        body = self.full_text
        original_post = str(self.original_post) if self.original_post else ''
        return '\n'.join((channel_post_url, header, body, original_post))

    def discord_embed(self) -> List[dict]:
        channel_url = f'https://www.youtube.com/channel/{self.channel_id}'
        post_url = f'https://www.youtube.com/post/{self.post_id}'

        original_post = str(self.original_post) if self.original_post else ''
        text = '\n'.join([self.full_text, '', original_post])

        embed = {
            'title': self.post_id,
            'description': text,
            'url': post_url,
            'color': None,
            'author': {'name': self.author, 'url': channel_url, 'icon_url': self.avatar_url},
            'footer': {'text': self.published_text}
        }
        embeds = [embed]
        return embeds


@Plugins.register('community', Plugins.kind.ACTOR_CONFIG)
class CommunityPostsMonitorConfig(PagedFeedMonitorConfig):
    pass


@Plugins.register('community', Plugins.kind.ACTOR_ENTITY)
class CommunityPostsMonitorEntity(PagedFeedMonitorEntity):
    url: str
    """url of the community page of the channel"""
    update_interval: float = 1800
    """how often the community page will be checked for new posts"""

@Plugins.register('community', Plugins.kind.ACTOR)
class CommunityPostsMonitor(PagedFeedMonitor):
    """
    Youtube community page monitor

    Monitors posts on community page of a channel, supports
    member-only posts if login cookies are provided. Some features,
    such as polls, are not supported.

    Examples of supported url:

    - `https://www.youtube.com/@ChannelName/community`
    - `https://www.youtube.com/channel/UCK0V3b23uJyU4N8eR_BR0QA/community`
    """

    def get_record_id(self, record: CommunityPostRecord) -> str:
        return f'{record.channel_id}:{record.post_id}'

    async def handle_first_page(self, entity: PagedFeedMonitorEntity, session: aiohttp.ClientSession) -> Tuple[Optional[Sequence[Record]], Optional[Any]]:
        raw_page_text = await self.request(entity.url, entity, session)
        if raw_page_text is None:
            return None, None
        raw_page_text = await handle_consent(raw_page_text, entity.url, session, self.logger)
        try:
            initial_page = get_initial_data(raw_page_text)
        except Exception as e:
            self.logger.exception(f'[{entity.name}] failed to get initial data from {entity.url}: {e}')
            return None, None
        records = self._parse_entries(initial_page)
        continuation_token = get_continuation_token(initial_page)
        innertube_context = get_innertube_context(raw_page_text)

        return records, (innertube_context, continuation_token)

    async def handle_next_page(self, entity: PagedFeedMonitorEntity, session: aiohttp.ClientSession, context: Optional[Any]) -> Tuple[Optional[Sequence[Record]], Optional[Any]]:
        innertube_context, continuation_token = context  # type: ignore
        if continuation_token is None:
            self.logger.debug(f'[{entity.name}] no continuation for next page, done loading')
            return [], None

        url, headers, post_body = prepare_next_page_request(innertube_context, continuation_token, cookies=session.cookie_jar)
        current_page = await utils.request_json(url, session, self.logger, method='POST', headers=headers,
                                                data=json.dumps(post_body), retry_times=3, retry_multiplier=2,
                                                retry_delay=5)
        if current_page is None:
            self.logger.debug(f'[{entity.name}] failed to load next page, aborting')
            return None, None
        current_page_records = self._parse_entries(current_page) or []
        continuation_token = get_continuation_token(current_page)
        context = (innertube_context, continuation_token) if continuation_token else None
        return current_page_records, context

    def _parse_entries(self, page: dict) -> List[Union[CommunityPostRecord, SharedCommunityPostRecord]]:
        renderers = get_renderers(page)
        records: List[Union[CommunityPostRecord, SharedCommunityPostRecord]] = []
        post_renderers = get_posts_renderers(renderers)
        for item in post_renderers:
            try:
                info = CommunityPostInfo.from_post_renderer(item)
                record = CommunityPostRecord(**info.model_dump())
                records.append(record)
            except Exception as e:
                self.logger.debug(f'error parsing post renderer {item}: {e}')
                continue
        shared_post_renderers = get_shared_posts_renderers(renderers)
        for item in shared_post_renderers:
            try:
                info = SharedCommunityPostInfo.from_post_renderer(item)
                record = SharedCommunityPostRecord(**info.model_dump())
                records.append(record)
            except Exception as e:
                self.logger.debug(f'error parsing shared post renderer {item}: {e}')
                continue
        return records




import datetime
import textwrap
from enum import Enum
from typing import List, Optional, Type

import markdown
from markdown.extensions.toc import TocExtension
from pydantic import BaseModel
from pydantic.fields import FieldInfo
from pydantic_core import PydanticUndefined

from avtdl.core.plugins import Plugins

HELP_FILE_STATIC_PART = f'''
<!-- This file is autogenerated as part of avtdl at {datetime.datetime.now()} -->

## Description and configuration of available plugins

[TOC]
'''


# implicitly relies on first of the lines inside `description` being a short title
PLUGIN_INFO_TEMPLATE = '''
### `{name}` - {description}
'''

PLUGIN_OPTIONS_TEMPLATE = '''
#### Plugin configuration options:
{config}

'''
ENTITY_OPTIONS_TEMPLATE = '''
#### Entity configuration options:
{entity}
'''

LIST_ITEM_TEMPLATE = '* `{name}`: {description}'

HTML_PAGE_TEMPLATE = '''
<!DOCTYPE html>
<html lang="en">
  <head>
    <meta charset="utf-8">
    <title>...</title>
    <link rel="stylesheet" href="modest.css">
  </head>
  <body>
    {body}
  </body>
</html>
'''

ASSOCIATED_RECORDS_TEMPLATE = '''
#### Produced records types:
'''

COLLAPSIBLE_ITEM_TEMPLATE = '''
<details markdown="block">
  <summary>{title}</summary>

{content}

</details>
'''

def get_plugin_info(plugin_name: str) -> str:
    plugin, config, entity = Plugins.get_actor_factories(plugin_name)
    description = render_doc(plugin)
    text = [PLUGIN_INFO_TEMPLATE.format(name=plugin_name, description=description)]
    config_info = get_config_model_info(config)
    if config_info:
        text.append(PLUGIN_OPTIONS_TEMPLATE.format(config=config_info))
    entity_info = get_entity_model_info(entity)
    if entity_info:
        text.append(ENTITY_OPTIONS_TEMPLATE.format(entity=entity_info))
    associated_records = Plugins.get_associated_records(plugin_name)
    if associated_records:
        records_text = [ASSOCIATED_RECORDS_TEMPLATE]
        for record_type in associated_records:
            record_info = get_record_model_info(record_type)
            record_text = COLLAPSIBLE_ITEM_TEMPLATE.format(title=record_type.__name__, content=record_info)
            records_text.append(record_text)
        text.extend(records_text)
    return '\n'.join(text)


def get_model_info(model: Type[BaseModel], skip_name: bool = False) -> str:
    info: List[str] = []
    description = render_doc(model)
    if description:
        info.append(description)
        info.append('\n') # ensure newline before list to make it render correctly
    required_fields = []
    not_required_fields = []
    for name, field_info in model.model_fields.items():
        if skip_name and name == 'name':
            continue
        if field_info.exclude:
            continue
        field_description = render_field_info(field_info, skip_details=False)
        field_description_text = LIST_ITEM_TEMPLATE.format(name=name, description=field_description)
        if has_default(field_info):
            not_required_fields.append(field_description_text)
        else:
            required_fields.append(field_description_text)
    info.extend(required_fields)
    if required_fields and not_required_fields:
        info.append('##### ') # separate mandatory and non-mandatory fields in two lists
    info.extend(not_required_fields)
    return '\n'.join(info)


def get_config_model_info(model: Type[BaseModel]) -> str:
    return get_model_info(model, skip_name=True)

def get_entity_model_info(model: Type[BaseModel]) -> str:
    return get_model_info(model, skip_name=False)


def get_record_model_info(model: Type[BaseModel]) -> str:
    info: List[str] = []
    description = render_doc(model)
    if description:
        info.append(description)
        info.append('\n') # ensure newline before list to make it render correctly
    for name, field_info in model.model_fields.items():
        if field_info.exclude:
            continue
        field_description = render_field_info(field_info, skip_details=True)
        field_description_text = LIST_ITEM_TEMPLATE.format(name=name, description=field_description)
        info.append(field_description_text)
    return '\n'.join(info)


def render_doc(model: Type[BaseModel]) -> str:
    if model.__doc__:
        text = textwrap.dedent(model.__doc__).strip('\n')
        return text
    return ''


def render_field_info(field_info: FieldInfo, skip_details=False) -> str:
    FIELD_INFO_TEMPLATE = '{description}. {details}'
    default = get_default(field_info)
    if skip_details:
        details = ''
    elif default:
        details = f'Default value is `{default}`.'
    else:
        details = 'Required.' if field_info.is_required() else 'Not required.'
    description = field_info.description or ''
    return FIELD_INFO_TEMPLATE.format(details=details, description=description)


def has_default(field_info: FieldInfo) -> bool:
    return field_info.default is not PydanticUndefined

def get_default(field_info: FieldInfo) -> Optional[str]:
    """Return text describing default value of given FieldInfo if set"""
    value = field_info.default
    if value is PydanticUndefined:
        return None
    if value is None:
        return None
    if isinstance(value, bool):
        return str(value).lower()
    if isinstance(value, Enum):
        return value.value
    if isinstance(value, str):
        if not value.strip(' \t\r\n'):
            return None
        return value
    if isinstance(value, dict):
        return ', '.join(f'"{k}": "{v}"' for k, v in value.items())
    if isinstance(value, type):
        return value.__name__
    return str(value)


def render_plugins_descriptions() -> str:
    """load available plugins and generate a help file in markdown from docstrings"""
    Plugins.load()
    descriptions = {name: get_plugin_info(name) for name in Plugins.known[Plugins.kind.ACTOR].keys()}
    SEPARATOR = '\n---\n'
    text = SEPARATOR.join(['', *descriptions.values(), ''])
    return HELP_FILE_STATIC_PART + text


def render_markdown(text: str) -> str:
    """convert markdown to html fragment"""
    md = markdown.Markdown(extensions=[TocExtension(toc_depth=3), 'md_in_html'])
    html = md.convert(text)
    return html


def generate_plugins_description(as_html: bool = False):
    text = render_plugins_descriptions()
    if not as_html:
        return text
    html = render_markdown(text)
    html = HTML_PAGE_TEMPLATE.format(body=html)
    return html



import datetime
import json
import re
import urllib.request
from collections import defaultdict
from json import JSONDecodeError
from typing import Any, Dict, List, Optional

import aiohttp
from pydantic import BaseModel, Field


class LoginRequiredError(ValueError):
    '''Raised when video page returns no data aside from {'playability_status': 'LOGIN_REQUIRED'}, which usually indicated video being private'''

class VideoErrorError(ValueError):
    '''Raised when video page returns {'playability_status': 'ERROR'}'''

class VideoFormat(BaseModel):
    itag: int
    url: str = Field(repr=False)
    mime: str
    quality: str

class VideoInfo(BaseModel):
    url: str
    title: str
    published: str
    uploaded: str
    author: str
    channel_id: str
    video_id: str
    summary: str = Field(repr=False)
    views: int
    length: int

    scheduled: Optional[datetime.datetime] = None
    live_start: Optional[datetime.datetime] = None
    live_end: Optional[datetime.datetime] = None

    is_unlisted: bool
    is_adult: bool
    is_livestream: bool
    is_upcoming: bool

    playability_status: str
    playability_reason: Optional[str] = None

    formats: Optional[List[VideoFormat]] = []


def get_video_page(url: str) -> str:
    with urllib.request.urlopen(url, timeout=15) as resp:
        data = resp.read()
        return data.decode('utf8')

async def aget_video_page(url: str, session: Optional[aiohttp.ClientSession] = None) -> str:
    if session is None:
        async with aiohttp.request('GET', url) as response:
            text = await response.text()
    else:
        async with session.get(url) as response:
            text = await response.text()
    return text

def get_initial_player_response(page: str) -> dict:
    try:
        return get_initial_response_fast(page)
    except (ValueError, JSONDecodeError):
        return get_initial_response_slow(page)

def get_initial_response_fast(page: str) -> dict:
    re_initial_data = 'var ytInitialPlayerResponse = ([^;]*);'
    match = re.search(re_initial_data, page)
    if match is None:
        raise ValueError(f'Failed to find ytInitialPlayerResponse on the page')
    raw_data = match.groups()[0]
    data = json.loads(raw_data)
    return data

def get_initial_response_slow(page: str) -> dict:
    anchor = 'var ytInitialPlayerResponse = {'
    pos_start = page.find(anchor)
    if pos_start == -1:
        raise ValueError(f'Failed to find ytInitialPlayerResponse on the page')
    pos_start += len(anchor) - 1
    position = pos_start

    re_parenthesses = re.compile('[{}]')
    parenthesses_values = defaultdict(int, {'{': 1, '}': -1})
    parentheses = 0
    while True:
        parentheses += parenthesses_values[page[position]]
        if parentheses == 0:
            raw_data = page[pos_start:position+1]
            response = json.loads(raw_data)
            return response
        position_match = re_parenthesses.search(page, position + 1)
        try:
            position = position_match.start()
        except AttributeError:
            raise ValueError(f'Failed to find matching set of parentheses after ytInitialPlayerResponse')

def get_embedded_player_response(page: str) -> dict:
    pos_start = page.find('{"embedded_player_response"')
    if pos_start == -1:
        raise ValueError(f'Failed to find embedded_player_response on page')
    parentheses = 1
    for position in range(pos_start+1, len(page)):
        if page[position] == '{':
            parentheses += 1
        if page[position] == '}':
            parentheses -= 1
        if parentheses == 0:
            raw_data = page[pos_start:position+1]
            response_data = json.loads(raw_data).get('embedded_player_response', '')
            response = json.loads(response_data)
            return response
    else:
        raise ValueError(f'Failed to find closing parenthesis for embedded_player_response')

def rename_keys(input_dict: Dict[str, Any], key_mapping: Dict[str, str]) -> Dict[str, Any]:
    output_dict = {}
    for data_key, input_key in key_mapping.items():
        value = input_dict.get(input_key)
        if value is not None:
            output_dict[data_key] = value
    return output_dict

def parse_playability_status(player_response: dict) -> Dict[str, Any]:
    playability_status = player_response.get('playabilityStatus')
    if playability_status is None:
        return {}

    info = rename_keys(playability_status, {'playability_status': 'status', 'playability_reason': 'reason'})

    try:
        date = playability_status["liveStreamability"]["liveStreamabilityRenderer"]["offlineSlate"]["liveStreamOfflineSlateRenderer"]["scheduledStartTime"]
        date = datetime.datetime.fromtimestamp(int(date), tz=datetime.timezone.utc)
    except (KeyError, TypeError, ValueError):
        pass
    else:
        info['scheduled'] = date

    return info

def parse_video_details(player_response: dict) -> Dict[str, Any]:
    video_details = player_response.get('videoDetails')
    if video_details is None:
        return {}
    key_mapping = {
        'title': 'title',
        'author': 'author',
        'channel_id': 'channelId',
        'video_id': 'videoId',
        'summary': 'shortDescription',
        'views': 'viewCount',
        'length': 'lengthSeconds',
        'is_livestream': 'isLiveContent',
    }
    info = rename_keys(video_details, key_mapping)
    info['is_upcoming'] = video_details.get('isUpcoming', False)
    return info

def parse_microformat(player_response: dict) -> Dict[str, Any]:
    microformat = player_response.get('microformat', {}).get('playerMicroformatRenderer')
    if microformat is None:
        return {}
    key_mapping = {
        'published': 'publishDate',
        'uploaded': 'uploadDate',
        'author': 'ownerChannelName',
        'views': 'viewCount',
        'length': 'lengthSeconds',
        'is_unlisted': 'isUnlisted',
    }
    info = rename_keys(microformat, key_mapping)
    if microformat.get('title') is not None:
        info['title'] = microformat.get('title', {}).get('simpleText')
    if microformat.get('description') is not None:
        info['summary'] = microformat.get('description', {}).get('simpleText')
    info['is_adult'] = not microformat.get('isFamilySafe')
    info['is_livestream'] = microformat.get('liveBroadcastDetails') is not None

    live_details = microformat.get('liveBroadcastDetails')
    if live_details is not None:
        if live_details.get('endTimestamp') is not None: # live ended
            info['live_start'] = live_details.get('startTimestamp')
            info['live_end'] = live_details.get('endTimestamp')
        elif live_details.get('isLiveNow') == True: # live is live
            info['live_start'] = live_details.get('startTimestamp')
        else: # live is scheduled
            date = live_details.get('startTimestamp')
            try:
                date = datetime.datetime.fromisoformat(date)
                info['scheduled'] = date
            except (ValueError, TypeError):
                pass

    return info

def parse_video_formats(player_response: dict) -> List[VideoFormat]:
    formats = player_response.get('streamingData')
    if formats is None:
        return []
    formats_list = []
    formats_list.extend(formats.get('formats', []))
    formats_list.extend(formats.get('adaptiveFormats', []))

    parsed_formats = []
    for item in formats_list:
        try:
            itag = item['itag']
            url = item['url']
            mime = item['mimeType']
            quality = item.get('qualityLabel') or item['quality']
            parsed_formats.append(VideoFormat(itag=itag, url=url, mime=mime, quality=quality))
        except (TypeError, KeyError):
            continue
    return parsed_formats


def parse_player_response(player_response: dict) -> Dict[str, Any]:
    info = {}
    info.update(parse_video_details(player_response))
    info.update(parse_microformat(player_response))
    info.update(parse_playability_status(player_response))
    info['formats'] = parse_video_formats(player_response)
    return info

def parse_video_page(page: str, url: str) -> VideoInfo:
    response = get_initial_player_response(page)
    data = parse_player_response(response)
    data['url'] = url
    if data['playability_status'] == 'LOGIN_REQUIRED' and data.get('playability_reason') is None:
        raise LoginRequiredError(f'Video is likely private: {url}. Raw data: {data}')
    if data['playability_status'] == 'ERROR':
        raise VideoErrorError(f'Video was deleted or never existed: {url} . Raw data: {data}')
    info = VideoInfo(**data)
    return info

def get_video_info(url: str) -> VideoInfo:
    page = get_video_page(url)
    return parse_video_page(page, url)

async def aget_video_info(url: str, session: Optional[aiohttp.ClientSession] = None) -> VideoInfo:
    page = await aget_video_page(url, session)
    return parse_video_page(page, url)

if __name__ == '__main__':
    from pprint import pprint
    urls = {
        'past live': 'https://www.youtube.com/watch?v=4edZ1YcVnhE',
        'future live': 'https://www.youtube.com/watch?v=rKMhl43RHo0',
        'ongoing live': 'https://www.youtube.com/watch?v=rUxyKA_-grg',
        'adult': 'https://www.youtube.com/watch?v=0UcBpNS8-P4',
        'premium': 'https://www.youtube.com/watch?v=sY1tXDR-oBg',
        'member only': 'https://www.youtube.com/watch?v=mcZ6dN_wGc8',
        'unlisted': 'https://www.youtube.com/live/IJXKVtb3xuU?si=o0gUYiYGoWW5_ukt',
        'not a live vod': 'https://www.youtube.com/watch?v=SdF0vjVj278',
        'privated': 'https://www.youtube.com/watch?v=w_ivzW_ya-U',
        'nonexistant': 'https://www.youtube.com/watch?v=mamamama',
    }
    for name, url in urls.items():
        try:
            info = get_video_info(url)
            info.summary = '0'
            info.formats = []
        except Exception as e:
            print(f'[****** {name} ********]')
            print(f'{e!r}')
        else:
            print(f'[****** {name} ********]')
            pprint(info.model_dump())
